import { apFirst as apFirst_, apS as apS_, apSecond as apSecond_, } from "fp-ts/lib/Apply";
import * as A from "fp-ts/lib/Array";
import * as chainable from "fp-ts/lib/Chain";
import { bindTo as bindTo_, flap as flap_, let as let_, } from "fp-ts/lib/Functor";
import * as O from "fp-ts/lib/Option";
import * as T from "fp-ts/lib/Task";
import * as TO from "fp-ts/lib/TaskOption";
import { constTrue, constant, flow, identity, pipe, } from "fp-ts/lib/function";
import * as AG from "./AsyncGenerator";
import { getAsyncIteratorNextTask, reduceUntilWithIndexLimited, } from "./AsyncIterableReduce";
import * as I from "./Iterable";
import { asUnit as asUnit_, as as as_, dual, tapIO as tapIO_, tapTask as tapTask_, tap as tap_, } from "./internal";
/**
 * @category type lambdas
 * @since 1.0.0
 */
export const URI = "AsyncIterable";
/**
 * Return a `AsyncIterable` with elements initialized with `f(i)`.
 *
 * Iterable stops when f return O.none
 *
 * @category constructors
 * @since 1.0.0
 */
export const makeByWithIndex = (f) => {
    let i = 0;
    return {
        async *[Symbol.asyncIterator]() {
            const el = f(i++);
            if (O.isSome(el)) {
                yield el.value;
            }
        },
    };
};
/**
 * @category constructors
 * @since 1.0.0
 */
export const makeBy = (f) => makeByWithIndex(() => f());
/**
 * Return a `AsyncIterable` with elements initialized with `f(i)`.
 *
 * Iterable stops when f return O.none
 *
 * @category constructors
 * @since 1.0.0
 */
export const makeByTaskWithIndex = (f) => {
    let i = 0;
    return {
        async *[Symbol.asyncIterator]() {
            const optionA = await f(i++)();
            if (O.isSome(optionA)) {
                yield optionA.value;
            }
        },
    };
};
/**
 * @category constructors
 * @since 1.0.0
 */
export const makeByTask = (f) => makeByTaskWithIndex(() => f());
/**
 * @category constructors
 * @since 1.0.0
 */
export const unfold = (b, f) => {
    let bb = b;
    return {
        async *[Symbol.asyncIterator]() {
            const mt = f(bb);
            if (O.isSome(mt)) {
                const [a, b] = mt.value;
                bb = b;
                yield a;
            }
        },
    };
};
/**
 * @category constructors
 * @since 1.0.0
 */
export const unfoldTask = (b, f) => {
    let bb = b;
    return {
        async *[Symbol.asyncIterator]() {
            const mt = await f(bb)();
            if (O.isSome(mt)) {
                const [a, b] = mt.value;
                bb = b;
                yield a;
            }
        },
    };
};
/**
 * @category conversions
 * @since 1.0.0
 */
export const fromIterable = (fa) => pipe(fa, AG.fromIterable, fromAsyncGenerator);
/**
 * @category conversions
 * @since 1.0.0
 */
export function fromAsyncGenerator(fa) {
    return {
        async *[Symbol.asyncIterator]() {
            yield* fa();
        },
    };
}
/**
 * @category conversions
 * @since 1.0.0
 */
export const fromLazyArg = AG.fromLazyArg;
/**
 * @category conversions
 * @since 1.0.0
 */
export const fromIO = fromLazyArg;
/**
 * @category conversions
 * @since 1.0.0
 */
export const fromTask = AG.fromTask;
/**
 * @category constructors
 * @since 1.0.0
 */
export const of = (a) => fromLazyArg(constant(a));
/**
 * @category instances
 * @since 1.0.0
 */
export const Pointed = { URI, of };
/**
 * @category mapping
 * @since 1.0.0
 */
export const mapWithIndex = (f) => (fa) => {
    let i = 0;
    return {
        async *[Symbol.asyncIterator]() {
            for await (const a of fa) {
                yield f(i++, a);
            }
        },
    };
};
/**
 * @category mapping
 * @since 1.0.0
 */
export const map = (f) => mapWithIndex((_i, a) => f(a));
/**
 * @category instances
 * @since 1.0.0
 */
export const Functor = {
    URI,
    map: (fa, f) => map(f)(fa),
};
/**
 * @category instances
 * @since 1.0.0
 */
export const FunctorWithIndex = {
    ...Functor,
    mapWithIndex: (fa, f) => mapWithIndex(f)(fa),
};
/**
 * Maps every value to the specified constant value.
 *
 * @category mapping
 * @since 1.0.0
 */
export const as = dual(2, as_(Functor));
/**
 * Maps every value to the void constant value.
 *
 * @category mapping
 * @since 1.0.0
 */
export const asUnit = asUnit_(Functor);
/**
 * @category mapping
 * @since 1.0.0
 */
export const flap = flap_(Functor);
/**
 * @category apply
 * @since 1.0.0
 */
export const ap = (fa) => (fab) => {
    const fabTask = pipe(fab, toArraySeq());
    let fabPromise;
    let fabs;
    let i = 0;
    let a;
    let justStarted = true;
    return {
        async *[Symbol.asyncIterator]() {
            if (!fabPromise) {
                fabPromise = fabTask();
            }
            if (!fabs) {
                fabs = await fabPromise;
            }
            if (i === fabs.length || justStarted) {
                i = 0;
                justStarted = false;
                for await (a of fa) {
                    yield fabs[i++](a);
                }
            }
            else {
                yield fabs[i++](a);
            }
        },
    };
};
/**
 * @category apply
 * @since 1.0.0
 */
export const apTask = (fa) => (fab) => {
    const fabTask = pipe(fab, toArraySeq());
    let fabPromise;
    let fabs;
    let i = 0;
    let a;
    let justStarted = true;
    return {
        async *[Symbol.asyncIterator]() {
            if (!fabPromise) {
                fabPromise = fabTask();
            }
            if (!fabs) {
                fabs = await fabPromise;
            }
            if (i === fabs.length || justStarted) {
                justStarted = false;
                i = 0;
                for await (a of fa) {
                    yield await fabs[i++](a)();
                }
            }
            else {
                yield await fabs[i++](a)();
            }
        },
    };
};
/**
 * @category instances
 * @since 1.0.0
 */
export const Apply = {
    ...Functor,
    ap: (fab, fa) => ap(fa)(fab),
};
/**
 * @category apply
 * @since 1.0.0
 */
export const apFirst = apFirst_(Apply);
/**
 * @category apply
 * @since 1.0.0
 */
export const apSecond = apSecond_(Apply);
/**
 * @category instances
 * @since 1.0.0
 */
export const Applicative = {
    ...Pointed,
    ...Apply,
};
/**
 * @category sequencing
 * @since 1.0.0
 */
export const flatMap = (f) => (fa) => {
    const buffer = [];
    const toBufferQueue = [];
    const fromBufferQueue = [];
    const nextAIFB = pipe(fa, getAsyncIteratorNextTask, TO.map(f));
    let isNextAIFBDone = false;
    let waitingForValues = 0; // fight against race conditions
    async function toBuffer() {
        if (!isNextAIFBDone && toBufferQueue.length === 0) {
            waitingForValues++;
            const maybeAIB = await nextAIFB();
            waitingForValues--;
            if (O.isSome(maybeAIB)) {
                const nextAIB = getAsyncIteratorNextTask(maybeAIB.value);
                toBufferQueue.push(nextAIB);
            }
            else {
                isNextAIFBDone = true;
            }
        }
        if (toBufferQueue.length === 0) {
            return;
        }
        const nextAIB = toBufferQueue[0];
        waitingForValues++;
        const maybeB = await nextAIB();
        waitingForValues--;
        if (O.isSome(maybeB)) {
            buffer.push(maybeB.value);
        }
        else {
            const i = toBufferQueue.findIndex((aib) => aib === nextAIB);
            if (i !== -1) {
                toBufferQueue.splice(i, 1);
            }
            if (!isNextAIFBDone ||
                (buffer.length === 0 && toBufferQueue.length > 0)) {
                await toBuffer();
            }
        }
        dispatchFromBuffer();
    }
    function dispatchFromBuffer() {
        if (buffer.length > 0) {
            const fromTasks = fromBufferQueue.splice(0, buffer.length);
            const values = buffer.splice(0, fromTasks.length);
            pipe(fromTasks, A.zip(values), A.map(([task, value]) => {
                task(O.some(value));
            }));
        }
        if (waitingForValues > 0 ||
            toBufferQueue.length > 0 ||
            buffer.length > 0) {
            return;
        }
        const hopeless = fromBufferQueue.splice(0);
        hopeless.forEach((task) => {
            task(O.none);
        });
    }
    function addToFromBufferQueue() {
        return new Promise((resolve) => fromBufferQueue.push(resolve));
    }
    async function next() {
        if (buffer.length > 0) {
            return O.some(buffer.shift());
        }
        const fromBuffer = addToFromBufferQueue();
        await toBuffer();
        return fromBuffer;
    }
    return {
        async *[Symbol.asyncIterator]() {
            const o = await next();
            if (O.isSome(o)) {
                yield o.value;
            }
        },
    };
};
/**
 * @category sequencing
 * @since 1.0.0
 */
export const flatMapIterable = (f) => (fa) => {
    const buffer = [];
    const fromBufferQueue = [];
    const nextAIFB = pipe(fa, getAsyncIteratorNextTask, TO.map(f));
    let isNextAIFBDone = false;
    let waitingForValues = 0; // fight against race conditions
    async function toBuffer() {
        if (!isNextAIFBDone) {
            waitingForValues++;
            const maybeAIB = await nextAIFB();
            waitingForValues--;
            if (O.isSome(maybeAIB)) {
                buffer.push(...maybeAIB.value);
            }
            else {
                isNextAIFBDone = true;
            }
        }
        if (!isNextAIFBDone && buffer.length === 0) {
            await toBuffer();
        }
        dispatchFromBuffer();
    }
    function dispatchFromBuffer() {
        if (buffer.length > 0) {
            const fromTasks = fromBufferQueue.splice(0, buffer.length);
            const values = buffer.splice(0, fromTasks.length);
            pipe(fromTasks, A.zip(values), A.map(([task, value]) => {
                task(O.some(value));
            }));
        }
        if (waitingForValues > 0 || buffer.length > 0) {
            return;
        }
        const hopeless = fromBufferQueue.splice(0);
        hopeless.forEach((task) => {
            task(O.none);
        });
    }
    function addToFromBufferQueue() {
        return new Promise((resolve) => fromBufferQueue.push(resolve));
    }
    async function next() {
        if (buffer.length > 0) {
            return O.some(buffer.shift());
        }
        const fromBuffer = addToFromBufferQueue();
        await toBuffer();
        return fromBuffer;
    }
    return {
        async *[Symbol.asyncIterator]() {
            const o = await next();
            if (O.isSome(o)) {
                yield o.value;
            }
        },
    };
};
/**
 * @category sequencing
 * @since 1.0.0
 */
export const flatMapTaskWithIndex = (f) => (fa) => {
    let i = 0;
    return {
        async *[Symbol.asyncIterator]() {
            for await (const a of fa) {
                yield await f(i++, a)();
            }
        },
    };
};
/**
 * @category sequencing
 * @since 1.0.0
 */
export const flatMapTask = (f) => flatMapTaskWithIndex((_i, a) => f(a));
/**
 * @category sequencing
 * @since 1.0.0
 */
export const flatten = /*#__PURE__*/ flatMap(identity);
/**
 * @category sequencing
 * @since 1.0.0
 */
export const flattenIterable = /*#__PURE__*/ flatMapIterable(identity);
/**
 * @category instances
 * @since 1.0.0
 */
export const Chain = {
    ...Apply,
    chain: (fab, fa) => flatMap(fa)(fab),
};
/**
 * @category instances
 * @since 1.0.0
 */
export const Unfoldable = {
    URI,
    unfold,
};
/**
 * @category instances
 * @since 1.0.0
 */
export const Monad = {
    ...Pointed,
    ...Chain,
};
/**
 * @category instances
 * @since 1.0.0
 */
export const FromIO = {
    URI,
    fromIO,
};
/**
 * @category instances
 * @since 1.0.0
 */
export const FromTask = {
    URI,
    fromIO,
    fromTask,
};
/**
 * @category instances
 * @since 1.0.0
 */
export const MonadIO = {
    ...Monad,
    fromIO,
};
/**
 * Same as [`filter`](#filter), but passing also the index to the iterating function.
 *
 * @category filtering
 * @since 1.0.0
 */
export const filterWithIndex = (predicateWithIndex) => (fa) => {
    let i = 0;
    async function* next() {
        for await (const a of fa) {
            if (predicateWithIndex(i++, a)) {
                yield a;
            }
            else {
                yield* next();
            }
        }
    }
    return fromAsyncGenerator(next);
};
/**
 * @category filtering
 * @since 1.0.0
 */
export const filter = (predicate) => filterWithIndex((_, a) => predicate(a));
/**
 * @category filtering
 * @since 1.0.0
 */
export const filterTaskWithIndex = (predicateWithIndex) => (fa) => {
    let i = 0;
    async function* next() {
        for await (const a of fa) {
            if (await predicateWithIndex(i++, a)()) {
                yield a;
            }
            else {
                yield* next();
            }
        }
    }
    return fromAsyncGenerator(next);
};
/**
 * @category filtering
 * @since 1.0.0
 */
export const filterTask = (predicate) => filterTaskWithIndex((_, a) => predicate(a));
/**
 * @category filtering
 * @since 1.0.0
 */
export const filterMapWithIndex = (f) => (fa) => {
    let i = 0;
    async function* next() {
        for await (const a of fa) {
            const optionB = f(i++, a);
            if (O.isSome(optionB)) {
                yield optionB.value;
            }
            else {
                yield* next();
            }
        }
    }
    return fromAsyncGenerator(next);
};
/**
 * @category filtering
 * @since 1.0.0
 */
export const filterMap = (f) => filterMapWithIndex((_, a) => f(a));
/**
 * @category filtering
 * @since 1.0.0
 */
export const filterMapTaskWithIndex = (f) => (fa) => {
    let i = 0;
    async function* next() {
        for await (const a of fa) {
            const optionB = await f(i++, a)();
            if (O.isSome(optionB)) {
                yield optionB.value;
            }
            else {
                yield* next();
            }
        }
    }
    return fromAsyncGenerator(next);
};
/**
 * @category filtering
 * @since 1.0.0
 */
export const filterMapTask = (f) => filterMapTaskWithIndex((_, a) => f(a));
/**
 * Compacts an AsyncIterable of `Option`s discarding the `None` values and
 * keeping the `Some` values. It returns a new array containing the values of
 * the `Some` options.
 * @category filtering
 * @since 1.0.0
 */
export const compact = 
/*#__PURE__*/ filterMap(identity);
/**
 * @category filtering
 * @since 1.0.0
 */
export const rights = (fa) => {
    async function* next() {
        for await (const a of fa) {
            if (a._tag === "Right") {
                yield a.right;
            }
            else {
                yield* next();
            }
        }
    }
    return fromAsyncGenerator(next);
};
/**
 * @category filtering
 * @since 1.0.0
 */
export const lefts = (fa) => {
    async function* next() {
        for await (const a of fa) {
            if (a._tag === "Left") {
                yield a.left;
            }
            else {
                yield* next();
            }
        }
    }
    return fromAsyncGenerator(next);
};
/**
 * Creates a new `AsyncIterable` removing duplicate elements, keeping the first occurrence of an element, based on a `Eq<A>`.
 * @category filtering
 * @since 1.0.0
 */
export const uniq = (E) => (fa) => {
    const uniques = [];
    async function* next() {
        for await (const a of fa) {
            if (uniques.every((o) => !E.equals(o, a))) {
                uniques.push(a);
                yield a;
            }
            else {
                yield* next();
            }
        }
    }
    return fromAsyncGenerator(next);
};
/**
 * @category mapping
 * @since 1.0.0
 */
export function transform(transform, flush) {
    return (fa) => {
        let isFlushed = false;
        async function* next() {
            for await (const a of fa) {
                const b = transform(a);
                if (O.isSome(b)) {
                    yield b.value;
                }
                else {
                    yield* next();
                }
            }
            if (!isFlushed && flush) {
                isFlushed = true;
                yield flush();
            }
        }
        return fromAsyncGenerator(next);
    };
}
/**
 * @category mapping
 * @since 1.0.0
 */
export function transformTask(transform, flush) {
    return (fa) => {
        let isFlushed = false;
        async function* next() {
            for await (const a of fa) {
                const b = await transform(a)();
                if (O.isSome(b)) {
                    yield b.value;
                }
                else {
                    yield* next();
                }
            }
            if (!isFlushed && flush) {
                isFlushed = true;
                yield await flush()();
            }
        }
        return fromAsyncGenerator(next);
    };
}
/**
 * Composes computations in sequence, using the return value of one computation to determine the next computation and
 * keeping only the result of the first.
 *
 * @category combinators
 * @since 1.0.0
 */
export const tap = /*#__PURE__*/ dual(2, tap_(Chain));
/**
 * @category combinators
 * @since 1.0.0
 */
export const tapTask = /*#__PURE__*/ dual(2, tapTask_(FromTask, Chain));
/**
 * @category combinators
 * @since 1.0.0
 */
export const tapIO = /*#__PURE__*/ dual(2, tapIO_(FromIO, Chain));
// -------------------------------------------------------------------------------------
// do notation
// -------------------------------------------------------------------------------------
/**
 * @category do notation
 * @since 1.0.0
 */
export const Do = /*#__PURE__*/ of({});
/**
 * @category do notation
 * @since 1.0.0
 */
export const bind = chainable.bind(Chain);
/**
 * @category do notation
 * @since 1.0.0
 */
export const bindTo = bindTo_(Functor);
const _let = let_(Functor);
export { 
/**
 * @category do notation
 * @since 1.0.0
 */
_let as let, };
/**
 * @category do notation
 * @since 1.0.0
 */
export const apS = /*#__PURE__*/ apS_(Apply);
/**
 * the order of results does not corresponds to order of async iterator elements
 * @category folding
 * @since 1.0.0
 */
export function toIterableLimited(limit) {
    return flow(reduceUntilWithIndexLimited(limit, constTrue, A.zero(), (_, b, a) => {
        b.push(a);
        return b;
    }), T.map(I.fromReadonlyArray));
}
/**
 * @category folding
 * @since 1.0.0
 */
export function toIterablePar(limit) {
    return toIterableLimited(limit);
}
/**
 * @category folding
 * @since 1.0.0
 */
export function toIterableSeq() {
    return toIterableLimited(1);
}
/**
 * preserves the order of elements coming from async iterator and corresponding results
 * @category folding
 * @since 1.0.0
 */
export function toArrayLimited(limit) {
    return reduceUntilWithIndexLimited(limit, constTrue, A.zero(), (i, b, a) => {
        b[i] = a;
        return b;
    });
}
/**
 * @category folding
 * @since 1.0.0
 */
export function toArrayPar(limit) {
    return toArrayLimited(limit);
}
/**
 * @category folding
 * @since 1.0.0
 */
export function toArraySeq() {
    return toArrayLimited(1);
}
/**
 * @category folding
 * @since 1.0.0
 */
export const foldMapWithIndexPar = (M, limit) => (f) => (fa) => pipe(fa, reduceUntilWithIndexLimited(limit, constTrue, M.empty, (i, b, a) => M.concat(b, f(i, a))));
/**
 * @category folding
 * @since 1.0.0
 */
export const foldMapPar = (M, limit) => {
    const foldMapWithIndexM = foldMapWithIndexPar(M, limit);
    return (f) => foldMapWithIndexM((_, a) => f(a));
};
/**
 * @category folding
 * @since 1.0.0
 */
export const foldMapSeq = (M) => {
    const foldMapWithIndexM = foldMapWithIndexPar(M, 1);
    return (f) => foldMapWithIndexM((_, a) => f(a));
};
