/**
 * @since 1.0.0
 */
import * as O from "fp-ts/lib/Option";
import { nextTick } from "node:process";
/**
 * @category sequencing
 * @since 1.0.0
 */
export function getAsyncIteratorNextTask(iterable) {
    // let i = 0;
    async function next() {
        for await (const value of iterable) {
            return O.some(value);
        }
        return O.none;
    }
    return next;
}
/**
 * generic async iterable reducer.
 *
 * - limit: how many values pull in parallel.
 * - until: predicate if the function should keep pulling values from the iterator
 * - b: reducer accumulator initial value.
 * - f: reducer function.
 *
 * @category sequencing
 * @since 1.0.0
 */
export function reduceUntilWithIndexLimited(limit, until, b, f) {
    return (iterable) => () => new Promise((resolve, reject) => {
        let running = 0;
        let out = b;
        let index = 0;
        let isDone = false;
        const next = getAsyncIteratorNextTask(iterable);
        function run() {
            if (isDone || // we done
                running === limit // the tasks queue is saturated
            ) {
                return;
            }
            running++;
            next()
                .then(didRun)
                .catch((cause) => {
                isDone = true;
                reject(Error("This should never have happened. Use AsyncIterableEither or AsyncIterableOption tryCatch function to handle errors", { cause }));
            });
        }
        function didRun(value) {
            if (O.isNone(value)) {
                isDone = true;
            }
            else {
                out = f(index++, out, value.value);
                if (!until(value.value)) {
                    isDone = true;
                }
            }
            running--;
            // keeps it safe from race conditions
            // due to some iterators resolve synchronously
            nextTick(() => {
                run();
                if (running === 0) {
                    resolve(out);
                }
            });
        }
        while (running < limit) {
            run();
        }
    });
}
